# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================"""
""" Compute ion concentrations consistent with general
Poisson-Nernst-Planck (PNP) equations.

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:
  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
"""
import logging, os, sys
import numpy as np
import scipy.constants as sc
# import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)

# Druecke nicht-linearen Teil der Transportgleichung (genauer, des Flusses) ueber
# Bernoulli-Funktionen
#
# $$ B(x) = \frac{x}{\exp(x)-1} $$
#
# aus. Damit wir in der Naehe von 0 nicht "in die Bredouille geraten", verwenden
# wir hier lieber die Taylorentwicklung. In der Literatur (Selbherr, S. Analysis
# and Simulation of Semiconductor Devices, Spriger 1984) wird eine noch
# aufwendigere stueckweise Definition empfohlen, allerdings werden wir im
# Folgenden sehen, dass unser Ansatz fuer dieses stationaere Problem genuegt.
def B(x):
    """Bernoulli function"""
    return np.where( np.abs(x) < 1e-9,
        1 - x/2 + x**2/12 - x**4/720, # Taylor
        x / ( np.exp(x) - 1 ) )

# "lazy" Ansatz for approximating Jacobian
def jacobian(f, x0, dx=np.NaN):
    """Naive way to construct N x N Jacobin Fij from N-valued function
    f of N-valued vector x0.

    Parameters
    ----------
    f : callable
        N-valued function of N-valued variable vector
    x0 : (N,) ndarray
        N-valued variable vector
    dx : float (default: np.nan)
        Jacobian built with finite difference scheme of spacing ``dx``.
        If ``np.nan``, then use machine precision.

    Returns
    -------
    F : (N,N) ndarray
        NxN-valued 2nd order finite difference scheme approximate of Jacobian
        convention: F_ij = dfidxj, where i are array rows, j are array columns
    """


    N = len(x0)
    # choose step as small as possible
    if np.isnan(dx).any():
        res = np.finfo('float64').resolution
        dx = np.abs(x0) * np.sqrt(res)
        dx[ dx < res ] = res

    if np.isscalar(dx):
        dx = np.ones(N) * dx

    F = np.zeros((N,N)) # Jacobian Fij

    # convention: dfi_dxj
    # i are rows, j are columns
    for j in range(N):
        dxj = np.zeros(N)
        dxj[j] = dx[j]

        F[:,j] =  (f(x0 + dxj) - f(x0 - dxj)) / (2.0*dxj[j])

    return F

class PoissonNernstPlanckSystem:
    """Describes and solves a 1D Poisson-Nernst-Planck system"""

    # properties "offer" the solution in physical units:
    @property
    def grid(self):
        return self.X*self.l_unit

    @property
    def potential(self):
        return self.uij*self.u_unit

    @property
    def concentration(self):
        return self.nij*self.c_unit

    @property
    def charge_density(self):
        return np.sum(self.F * self.concentration.T * self.z,axis=1)

    #TODO:  replace "didactic" Newton solver from IMTEK Simulation course with
    #       some standard package
    def newton(self,f,xij):
        """Newton solver expects system f and initial value xij

        Parameters
        ----------
        f : callable
            N-valued function of N-valued vector
        xij : (N,) ndarray
            N-valued initial value vector

        Returns
        -------
        xij : (N,) ndarray
            N-valued solution vector
        """
        self.xij = []
        self.converged = True
        # assume convergence, set to 'false' if maxit exceeded later

        self.logger.debug('Newton solver, grid points N = {:d}'.format(self.N))
        self.logger.debug('Newton solver, tolerance e = {:> 8.4g}'.format(self.e))
        self.logger.debug('Newton solver, maximum number of iterations M = {:d}'.format(self.maxit))

        i = 0
        delta_rel = 2*self.e

        # if self.output:
        #     fig = plt.figure(figsize=(16,15))
        #     ax1 = plt.subplot(221)
        #     ax1.set_ylabel('u')
        #     ax1.set_xlabel('x')
        #     ax1.set_title('potential u')
        #
        #     ax2 = plt.subplot(322)
        #     ax2.set_ylabel('n')
        #     ax2.set_xlabel('x')
        #     ax2.set_title('density n')
        #
        #     ax3 = plt.subplot(325)
        #     ax3.set_ylabel('du')
        #     ax3.set_xlabel('x')
        #     ax3.set_title('step du')
        #
        #     ax4 = plt.subplot(326)
        #     ax4.set_ylabel('dn')
        #     ax4.set_xlabel('x')
        #     ax4.set_title('step dn')

        self.logger.info("Convergence criterion: norm(dx) < {:4.2e}".format(self.e))

        self.convergenceStepAbsolute = np.zeros(self.maxit)
        self.convergenceStepRelative = np.zeros(self.maxit)
        self.convergenceResidualAbsolute = np.zeros(self.maxit)

        dxij = np.zeros(self.N)
        while delta_rel > self.e and i < self.maxit:
            self.logger.debug('*** Newton solver iteration {:d} ***'.format(i))

            # avoid cluttering log
            self.logger.disabled = True
            J = jacobian(f, xij)
            self.logger.disabled = False

            rank = np.linalg.matrix_rank(J)
            self.logger.debug('    Jacobian ({}) rank {:d}'.format(J.shape, rank))

            if rank < self.N:
                self.logger.warn("Singular jacobian of rank"
                      + "{:d} < {:d} at step {:d}".format(
                      rank, self.N, i ))
                break

            F = f(xij)
            invJ = np.linalg.inv(J)

            dxij = np.dot( invJ, F )

            delta = np.linalg.norm(dxij)
            delta_rel = delta / np.linalg.norm(xij)

            xij -= dxij
            self.xij.append(xij)

            normF = np.linalg.norm(F)

            self.logger.debug('    convergence norm(dxij), absolute {:> 8.4g}'.format(delta))
            self.logger.debug('    convergence norm(dxij), realtive {:> 8.4g}'.format(delta_rel))
            self.logger.debug('          residual norm(F), absolute {:> 8.4g}'.format(normF))

            self.convergenceStepAbsolute[i] = delta
            self.convergenceStepRelative[i] = delta_rel
            self.convergenceResidualAbsolute[i] = normF
            self.logger.info("Step {:4d}: norm(dx)/norm(x) = {:4.2e}, norm(dx) = {:4.2e}, norm(F) = {:4.2e}".format(
                i, delta_rel, delta, normF) )
            # if i % self.outfreq == 0 and self.output:
            #
            #     duij = dxij[:self.Ni]
            #     dnij = dxij[self.Ni:(self.M+1)*self.Ni].reshape(self.M,self.Ni)
            #
            #     uij = xij[:self.Ni]
            #     nij = xij[self.Ni:(self.M+1)*self.Ni].reshape(self.M,self.Ni)
            #
            #     ax1.plot(self.X, uij, '-', label='u step {:02d}'.format(i))
            #     for k in range(self.M):
            #       ax2.plot(self.X, nij[k,:], '-', label='n step {:02d}, species {:02d}'.format(i,k))
            #     ax3.plot(self.X, duij, '-', label='du step {:02d}'.format(i))
            #     for k in range(self.M):
            #       ax4.plot(self.X, dnij[k,:], '-', label='dn step {:02d}, species {:02d}'.format(i,k))
            #
            #     ax1.legend(loc='best')
            #     ax2.legend(loc='best')
            #     ax3.legend(loc='best')
            #     ax4.legend(loc='best')

            i += 1

        if i == self.maxit:
            self.logger.warn("Maximum number of iterations reached")
            self.converged = False

        self.logger.info("Ended after {:d} steps.".format(i))
        # if self.output:
        #     fig = plt.figure(figsize=(16,10))
        #     ax1 = plt.subplot(221)
        #     ax1.set_ylabel(r'$\epsilon = \frac{|x_j - x_{j-1}|}{|x_{j-1}}$')
        #     ax1.set_xlabel('j')
        #     ax1.set_title('step convergence, relative')
        #     ax2 = plt.subplot(222)
        #     ax2.set_ylabel(r'\epsilon = $|x_j - x_{j-1}|')
        #     ax2.set_xlabel('j')
        #     ax2.set_title('step convergence, absolute')
        #     ax3 = plt.subplot(223)
        #     ax3.set_ylabel(r'$R = F(x_j)$')
        #     ax3.set_xlabel('j')
        #     ax3.set_title('residue convergence, absolute')
        #     ax1.plot(self.convergenceStepRelative[:i])
        #     ax2.plot(self.convergenceStepAbsolute[:i])
        #     ax3.plot(self.convergenceResidualAbsolute[:i])

        self.convergenceStepAbsolute = self.convergenceStepAbsolute[:i]
        self.convergenceStepRelative = self.convergenceStepRelative[:i]
        self.convergenceResidualAbsolute = self.convergenceResidualAbsolute[:i]
        return xij

    def init(self):
        """Sets up discretization scheme and initial value"""
        # indices
        self.Ni = self.N+1
        I = np.arange(self.Ni)

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'discretization segments N', self.N, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'grid points N', self.Ni, lwidth=self.label_width))

        # discretization
        self.dx      = self.L_scaled / self.N # spatial step
        # maximum time step allowed
        # (irrelevant for our steady state case)
        # D * dt / dx^2 <= 1/2
        # dt <= dx^2 / ( 2*D )
        # dt_max = dx**2 / 2 # since D = 1
        # dt_max = dx**2 / (2*self.Dn_scaled)

        # dx2overtau = dx**2 / self.tau_scaled
        self.dx2overtau = 10.0

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'dx', self.dx, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'dx2overtau', self.dx2overtau, lwidth=self.label_width))

        # positions (scaled)
        self.X = self.x0_scaled + I*self.dx

        # Bounary & initial values

        # internally:
        #   n: dimensionless concentrations
        #   u: dimensionless potential
        #   i: spatial index
        #   j: temporal index
        #   k: species index
        # initial concentrations equal to bulk concentrations

        # Kronecker product, M rows (ion species), Ni cols (grid points),
        self.ni0 = np.kron( self.c_scaled, np.ones((self.Ni,1)) ).T
        self.zi0 = np.kron( self.z, np.ones((self.Ni,1)) ).T # does not change

        self.initial_values()

    def initial_values(self):
        """
        Solves decoupled linear system to get inital potential distribution.
        """

        zini0 = self.zi0*self.ni0 # z*n
        #rhoi0 = np.zeros( self.Ni ) # initial charge distribution (dimensionless)
        #for k in range(0,self.M):
        #  rhoi0 += zini0[(k*self.Ni):((k+1)*self.Ni)]
        # shape: ion species (rows), grid points (cols), sum over ion species (along rows)
        rhoi0 = 0.5*zini0.sum(axis=0)

        # system matrix of spatial poisson equation
        Au = np.zeros((self.Ni,self.Ni))
        bu = np.zeros(self.Ni)
        Au[0,0]   = 1
        Au[-1,-1] = 1
        for i in range(1,self.N):
            Au[i,[i-1,i,i+1]] = [1.0, -2.0, 1.0] # 1D Laplace operator, 2nd order

        bu = rhoi0*self.dx**2 # => Poisson equation
        bu[0]  = self.u0
        bu[-1] = self.u1

        # get initial potential distribution by solving Poisson equation
        self.ui0 = np.dot( np.linalg.inv(Au), bu) # A u - b = 0 <=> u = A^-1 b

        return self.ui0

    # evokes Newton solver
    def solve(self):
        """Evokes newton solver

        Returns
        -------
        uij : (Ni,) ndarray
            potential at Ni grid points
        nij : (M,Nij) ndarray
            concentrations of M species at Ni grid points
        lamj: (L,) ndarray
            value of L Lagrange multipliers (not implemented, empty)
        """

        if len(self.g) > 0:
          self.xi0 = np.concatenate([self.ui0, self.ni0.flatten(), np.zeros(len(self.g))])
        else:
          self.xi0 = np.concatenate([self.ui0, self.ni0.flatten()])

        self.xij1 = self.newton(self.G,self.xi0.copy())

        # store results:
        self.uij  = self.xij1[:self.Ni] # potential
        self.nij  = self.xij1[self.Ni:(self.M+1)*self.Ni].reshape(self.M, self.Ni) # concentrations
        self.lamj = self.xij1[(self.M+1)*self.Ni:] # Lagrange multipliers

        return self.uij, self.nij, self.lamj

    # standard sets of boundary conditions:
    def useStandardInterfaceBC(self):
        """Interface at left hand side and open bulk at right hand side"""
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled
        self.u1 = 0
        self.logger.info('Left hand side Dirichlet boundary condition:                               u0 = {:> 8.4g}'.format(self.u0))
        self.logger.info('Right hand side Dirichlet boundary condition:                              u1 = {:> 8.4g}'.format(self.u1))

        self.boundary_conditions.extend([
          lambda x: self.leftPotentialDirichletBC(x,self.u0),
          lambda x: self.rightPotentialDirichletBC(x,self.u1) ])
        # self.rightPotentialBC = lambda x: self.rightPotentialDirichletBC(x,self.u1)

        #self.rightConcentrationBC = []
        for k in range(self.M):
          self.logger.info('Ion species {:02d} left hand side concentration Flux boundary condition:       j0 = {:> 8.4g}'.format(k,0))
          self.logger.info('Ion species {:02d} right hand side concentration Dirichlet boundary condition: c1 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.boundary_conditions.extend( [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) ] )
          #self.rightConcentrationBC.append(
          #  lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) )
        # counter-intuitive behavior of lambda in loop:
        # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
        # workaround: default parameter k=k

    def useStandardCellBC(self):
        """Interfaces at left hand side and right hand side"""
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.0
        self.u1 = - self.delta_u_scaled / 2.
        self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
          'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
        self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
          'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))
        self.boundary_conditions.extend([
          lambda x: self.leftPotentialDirichletBC(x,self.u0),
          lambda x: self.rightPotentialDirichletBC(x,self.u1) ])

        N0 = self.L_scaled*self.c_scaled # total amount of species in cell
        for k in range(self.M):
          self.logger.info('{:>{lwidth}s} j0 = {:<8.4g}'.format(
            'Ion species {:02d} left hand side concentration Flux boundary condition'.format(k),
            0.0, lwidth=self.label_width))
          self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
            'Ion species {:02d} number conservation constraint'.format(k),
            N0[k], lwidth=self.label_width))

          self.boundary_conditions.extend(  [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k, N0=N0[k]: self.numberConservationConstraint(x,k,N0) ] )

    def useSternLayerCellBC(self, implicit=False):
        """Interfaces at left hand side and right hand side,
        Stern layer either by prescirbing linear potential regime between cell
        boundary and outer Helmholtz plane (PHP), or by cropping the domain at
        OHP and applying Robin BC; zero flux BC on all ion species.

        Parameters
        ----------
        implicit : bool, optional
            If true, then computation domain is cropped by
            twice the Stern layer thickness lambda_S and true Robin BC
            are applied. Otherwise, enforces constant potential gradient across
            Stern layer region of thickness lambda_S. (default:False)
        """
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.0
        self.u1 = - self.delta_u_scaled / 2.0

        if implicit: # implicitly treat Stern layer via Robin BC
            self.logger.info('Implicitly treating Stern layer via Robin BC')

            self.L_scaled        = self.L_scaled - 2*self.lambda_S_scaled
            self.x0_scaled       = self.x0_scaled + self.lambda_S_scaled

            # no left and right hand side outer Helmholtz plane
            # self.lhs_ohp = self.x0_scaled
            # self.rhs_ohp = self.x0_scaled + self.L_scaled

            # print scaled quantities to log
            self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
              'reduced cropped domain size L*', self.L_scaled, lwidth=self.label_width))
            self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
              'shifted reduced reference position x0*', self.x0_scaled, lwidth=self.label_width))

            self.logger.info('{:>{lwidth}s} u0 + lambda_S*dudx = {:< 8.4g}'.format(
              'Left hand side Robin boundary condition', self.u0, lwidth=self.label_width))
            self.logger.info('{:>{lwidth}s} u1 + lambda_S*dudx = {:< 8.4g}'.format(
              'Right hand side Robin boundary condition', self.u1, lwidth=self.label_width))
            self.boundary_conditions.extend([
              lambda x: self.leftPotentialRobinBC(x,self.lambda_S_scaled,self.u0),
              lambda x: self.rightPotentialRobinBC(x,self.lambda_S_scaled,self.u1) ])

        else: # explicitly treat Stern layer via linear regime
            self.logger.info('Explicitly treating Stern layer as uniformly charged regions')

            # set eft and right hand side outer Helmholtz plane
            self.lhs_ohp = self.x0_scaled + self.lambda_S_scaled
            self.rhs_ohp = self.x0_scaled + self.L_scaled - self.lambda_S_scaled

            self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
              'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
            self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
              'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))
            self.boundary_conditions.extend([
              lambda x: self.leftPotentialDirichletBC(x,self.u0),
              lambda x: self.rightPotentialDirichletBC(x,self.u1) ])


        N0 = self.L_scaled*self.c_scaled # total amount of species in cell
        for k in range(self.M):
          self.logger.info('{:>{lwidth}s} j0 = {:<8.4g}'.format(
            'Ion species {:02d} left hand side concentration Flux boundary condition'.format(k),
            0.0, lwidth=self.label_width))
          self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
            'Ion species {:02d} number conservation constraint'.format(k),
            N0[k], lwidth=self.label_width))

          self.boundary_conditions.extend(  [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k, N0=N0[k]: self.numberConservationConstraint(x,k,N0) ] )

    # TODO: meaningful test for Dirichlet BC
    def useStandardDirichletBC(self):
        """Dirichlet BC for all variables at all boundaries"""
        self.boundary_conditions = []

        self.u0 = self.delta_u_scaled
        self.u1 = 0

        self.logger.info('Left hand side potential Dirichlet boundary condition:                     u0 = {:> 8.4g}'.format(self.u0))
        self.logger.info('Right hand side potential Dirichlet boundary condition:                    u1 = {:> 8.4g}'.format(self.u1))

        # set up boundary conditions
        self.boundary_conditions.extend( [
          lambda x: self.leftPotentialDirichletBC(x,self.u0),
          lambda x: self.rightPotentialDirichletBC(x,self.u1) ] )

        for k in range(self.M):
          self.logger.info('Ion species {:02d} left hand side concentration Dirichlet boundary condition:  c0 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.logger.info('Ion species {:02d} right hand side concentration Dirichlet boundary condition: c1 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.boundary_conditions.extend( [
            lambda x, k=k: self.leftDirichletBC(x,k,self.c_scaled[k]),
            lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) ] )

    # boundary conditions and constraints building blocks:
    def leftFiniteDifferenceSchemeFluxBC(self,x,k,j0=0):
        """
        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        j0 : float
            flux of ion species `k` at left hand boundary

        Returns
        -------
        float: boundary condition residual
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        # 2nd order right hand side finite difference scheme:
        # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2)
        # - dndx - z n dudx = j0
        dndx = -3.0*nijk[0] + 4.0*nijk[1] - nijk[2]
        dudx = -3.0*uij[0]  + 4.0*uij[1]  - uij[2]
        bcval = - dndx - self.zi0[k,0]*nijk[0]*dudx - 2.0*self.dx*j0

        self.logger.debug(
            'Flux BC F[0]  = - dndx - z n dudx - 2*dx*j0 = {:> 8.4g}'.format(bcval))
        self.logger.debug(
            '   = - ({:.2f}) - ({:.0f})*{:.2f}*({:.2f}) - 2*{:.2f}*({:.2f})'.format(
                dndx, self.zi0[k,0], nijk[0], dudx, self.dx, j0))
        return bcval

    def rightFiniteDifferenceSchemeFluxBC(self,x,k,j0=0):
        """
        See ```leftFiniteDifferenceSchemeFluxBC```
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        # 2nd order left hand side finite difference scheme:
        # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2)
        # - dndx - z n dudx = j0
        dndx = 3.0*nijk[-1] - 4.0*nijk[-2] + nijk[-3]
        dudx = 3.0*uij[-1]  - 4.0*uij[-2]  + uij[-3]
        bcval = - dndx - self.zi0[k,-1]*nijk[-1]*dudx - 2.0*self.dx*j0

        self.logger.debug(
            'FD flux BC F[-1]  = - dndx - z n dudx - 2*dx*j0 = {:> 8.4g}'.format(bcval))
        self.logger.debug(
            '  = - {:.2f} - {:.0f}*{:.2f}*{:.2f} - 2*{:.2f}*{:.2f}'.format(
                dndx, self.zi0[k,-1], nijk[-1], dudx, self.dx, j0))
        return bcval

    def leftControlledVolumeSchemeFluxBC(self,x,k,j0=0):
        """
        Compute left hand side flux boundary condition residual in accord with
        controlled volume scheme.

        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        j0 : float
            flux of ion species `k` at left hand boundary

        Returns
        -------
        float: boundary condition residual
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        # flux by controlled volume scheme:

        bcval = ( + B(self.z[k]*(uij[0]-uij[1]))*nijk[1]
                  - B(self.z[k]*(uij[1]-uij[0]))*nijk[0] - self.dx*j0 )

        self.logger.debug(
            'CV flux BC F[0]  = n1*B(z(u0-u1)) - n0*B(z(u1-u0)) - j0*dx = {:> 8.4g}'.format(bcval))
        return bcval

    def rightControlledVolumeSchemeFluxBC(self,x,k,j0=0):
        """
        Compute right hand side flux boundary condition residual in accord with
        controlled volume scheme. See ``leftControlledVolumeSchemeFluxBC``
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        # flux by controlled volume scheme:

        bcval = ( + B(self.z[k]*(uij[-2]-uij[-1]))*nijk[-1]
                  - B(self.z[k]*(uij[-1]-uij[-2]))*nijk[-2] - self.dx*j0 )

        self.logger.debug(
            'CV flux BC F[-1]  = n[-1]*B(z(u[-2]-u[-1])) - n[-2]*B(z(u[-1]-u[-2])) - j0*dx = {:> 8.4g}'.format(bcval))
        return bcval

    def leftPotentialDirichletBC(self,x,u0=0):
        return self.leftDirichletBC(x,-1,u0)

    def leftDirichletBC(self,x,k,x0=0):
        """Construct Dirichlet BC at left boundary"""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[0] - x0

    def rightPotentialDirichletBC(self,x,x0=0):
        return self.rightDirichletBC(x,-1,x0)

    def rightDirichletBC(self,x,k,x0=0):
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[-1] - x0

    def leftPotentialRobinBC(self,x,lam,u0=0):
        return self.leftRobinBC(x,-1,lam,u0)

    def leftRobinBC(self,x,k,lam,x0=0):
        """
        Compute left hand side Robin (u + lam*dudx = u0 ) BC at in accord with
        2nd order finite difference scheme.

        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        lam: float
            BC coefficient, corresponds to Stern layer thickness
            if applied to potential variable in PNP problem. Here, this steric
            layer is assumed to constitute a region of uniform charge density
            and thus linear potential drop across the interface.
        x0 : float
            right hand side value of BC, corresponds to potential beyond Stern
            layer if applied to poential variable in PNP system.

        Returns
        -------
        float: boundary condition residual
        """
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[0] + lam/(2*self.dx)* ( 3.0*nijk[0] - 4.0*nijk[1] + nijk[2] ) - x0

    def rightPotentialRobinBC(self,x,lam,u0=0):
        return self.rightRobinBC(x,-1,lam,u0)

    def rightRobinBC(self,x,k,lam,x0=0):
        """Construct Robin (u + lam*dudx = u0 ) BC at right boundary."""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[-1] + lam/(2*self.dx) * ( 3.0*nijk[-1] - 4.0*nijk[-2] + nijk[-3] ) - x0

    def numberConservationConstraint(self,x,k,N0):
        """N0: total amount of species, k: ion species"""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        # rescale to fit interval
        N = np.sum(nijk*self.dx) * self.N / self.Ni
        constraint_val = N - N0

        self.logger.debug(
            'Number conservation constraint F(x)  = N - N0 = {:.4g} - {:.4g} = {:.4g}'.format(
                N, N0, constraint_val ) )
        return constraint_val

    # TODO: remove or standardize
    # def leftNeumannBC(self,x,j0):
    #   """Construct finite difference Neumann BC (flux BC) at left boundary"""
    #   # right hand side first derivative of second order error
    #   # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2) = j0
    #   bcval = -3.0*x[0] + 4.0*x[1] - x[2] - 2.0*self.dx*j0
    #   self.logger.debug(
    #     'Neumann BC F[0]  = -3*x[0]  + 4*x[1]  - x[2]  = {:> 8.4g}'.format(bcval))
    #   return bcval
    #
    # def rightNeumannBC(self,x,j0):
    #   """Construct finite difference Neumann BC (flux BC) at right boundray"""
    #   # left hand side first derivative of second order error
    #   # dfndx = 1 / (2*dx) * (+3 fn - 4 fn-1 + fn-2 ) + O(dx^2) = 0
    #   bcval = 3.0*x[-1] - 4.0*x[-2] + x[-3] - 2.0*self.dx*j0
    #   self.logger.debug(
    #     'Neumann BC F[-1] = -3*x[-1] + 4*x[-2] - nijk[-3] = {:> 8.4g}'.format(bcval))
    #   return bcval

    # standard Poisson equation residual for potential
    def poisson_pde(self,x):
        """Returns Poisson equation resiudal by applying 2nd order FD scheme"""
        uij1 = x[:self.Ni]
        self.logger.debug(
          'potential range [u_min, u_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(uij1),np.max(uij1)))

        nij1 = x[self.Ni:(self.M+1)*self.Ni]

        nijk1 = nij1.reshape( self.M, self.Ni )
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} concentration range [c_min, c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(nijk1[k,:]),np.max(nijk1[k,:]) ) )

        # M rows (ion species), N_i cols (grid points)
        zi0nijk1 = self.zi0*nijk1 # z_ik*n_ijk
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} charge range [z*c_min, z*c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(zi0nijk1[k,:]),np.max(zi0nijk1[k,:]) ) )

        # charge density sum_k=1^M (z_ik*n_ijk)
        rhoij1 = zi0nijk1.sum(axis=0)
        self.logger.debug(
          'charge density range [rho_min, rho_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(rhoij1),np.max(rhoij1) ) )

        # reduced Poisson equation: d2udx2 = rho
        Fu = - ( np.roll(uij1, -1) - 2*uij1 + np.roll(uij1, 1) ) - 0.5 * rhoij1*self.dx**2

        # linear potential regime due to steric effects incorporated here
        # TODO: incorporate "spatially finite" BC into Robin BC functions
        # replace left and right hand side residuals with linear potential FD

        if not np.isnan(self.lhs_ohp):
            lhs_linear_regime_ndx = (self.X <= self.lhs_ohp)
            lhs_ohp_ndx = np.max( np.nonzero( lhs_linear_regime_ndx ) )

            self.logger.debug(
              'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled <= {:>.4g}'.format(
                np.count_nonzero(lhs_linear_regime_ndx), lhs_ohp_ndx, self.lhs_ohp) )

            # dudx = (u[ohp]-u[0])/lambda_S within Stern layer
            Fu[lhs_linear_regime_ndx] = (
                ( np.roll(uij1,-1) - uij1 )[lhs_linear_regime_ndx]
                    * self.lambda_S_scaled - (uij1[lhs_ohp_ndx]-uij1[0])*self.dx )

        if not np.isnan(self.rhs_ohp):
            rhs_linear_regime_ndx = (self.X >= self.rhs_ohp)
            rhs_ohp_ndx = np.min( np.nonzero( rhs_linear_regime_ndx ) )

            self.logger.debug(
              'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled >= {:>.4g}'.format(
                 np.count_nonzero(rhs_linear_regime_ndx), rhs_ohp_ndx, self.rhs_ohp) )

            # dudx = (u[ohp]-u[0])/lambda_S within Stern layer
            Fu[rhs_linear_regime_ndx] = (
                ( uij1 - np.roll(uij1,1) )[rhs_linear_regime_ndx]
                    * self.lambda_S_scaled - (uij1[-1]-uij1[rhs_ohp_ndx])*self.dx )

        Fu[0] = self.boundary_conditions[0](x)
        Fu[-1] = self.boundary_conditions[1](x)

        self.logger.debug('Potential BC residual Fu[0]  = {:> 8.4g}'.format(Fu[0]))
        self.logger.debug('Potential BC residual Fu[-1] = {:> 8.4g}'.format(Fu[-1]))

        return Fu

    def nernst_planck_pde(self,x):
        """Returns Nernst-Planck equation resiudal by applying controlled
        volume scheme"""
        uij1 = x[:self.Ni]

        self.logger.debug(
          'potential range [u_min, u_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(uij1),np.max(uij1)))

        nij1 = x[self.Ni:(self.M+1)*self.Ni]
        nijk1 = nij1.reshape( self.M, self.Ni )
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} concentration range [c_min, c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(nijk1[k,:]),np.max(nijk1[k,:]) ) )

        Fn = np.zeros([self.M, self.Ni])
        # loop over k = 1..M reduced Nernst-Planck equations:
        # - d2nkdx2 - ddx (zk nk dudx ) = 0
        for k in range(self.M):
          # conrolled volume implementation: constant flux across domain
          Fn[k,:] = (
            + B(self.zi0[k,:]*(uij1 - np.roll(uij1,-1))) * np.roll(nijk1[k,:],-1)
            - B(self.zi0[k,:]*(np.roll(uij1,-1) - uij1)) * nijk1[k,:]
            - B(self.zi0[k,:]*(np.roll(uij1,+1) - uij1)) * nijk1[k,:]
            + B(self.zi0[k,:]*(uij1 - np.roll(uij1,+1))) * np.roll(nijk1[k,:],+1) )

          # controlled volume implementation: flux j = 0 in every grid point
          #
          # Fn[k,:] =  (
          #   B(self.zi0[k,:]*(uij1 - np.roll(uij1,-1)))*np.roll(nijk1[k,:],-1)
          #   - B(self.zi0[k,:]*(np.roll(uij1,-1) - uij1))*nijk1[k,:] )

          # linear potential regime due to steric effects incorporated here
          # TODO: incorporate "spatially finite" BC into Robin BC functions
          # replace left and right hand side residuals with linear potential FD

          # left and right hand side outer Helmholtz plane
          # lhs_ohp = self.x0_scaled + self.lambda_S_scaled
          # rhs_ohp = self.x0_scaled + self.L_scaled - self.lambda_S_scaled
          #
          # lhs_linear_regime_ndx = (self.X <= lhs_ohp)
          # rhs_linear_regime_ndx = (self.X >= rhs_ohp)
          #
          # lhs_ohp_ndx = np.max( np.nonzero( lhs_linear_regime_ndx ) )
          # rhs_ohp_ndx = np.min( np.nonzero( rhs_linear_regime_ndx ) )
          #
          # self.logger.debug(
          #   'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled <= {:>.4g}'.format(
          #     np.count_nonzero(lhs_linear_regime_ndx), lhs_ohp_ndx, lhs_ohp) )
          # self.logger.debug(
          #   'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled >= {:>.4g}'.format(
          #      np.count_nonzero(rhs_linear_regime_ndx), rhs_ohp_ndx, rhs_ohp) )
          #
          # # zero concentration gradient in Stern layer
          # Fn[k,lhs_linear_regime_ndx] = (
          #     ( np.roll(nijk1[k,:],-1)-np.roll(nijk1[k,:],1))[lhs_linear_regime_ndx])
          # Fn[k,rhs_linear_regime_ndx] = (
          #     ( np.roll(nijk1[k,:],-1)-np.roll(nijk1[k,:],1))[rhs_linear_regime_ndx])

          Fn[k,0]  = self.boundary_conditions[2*k+2](x)
          Fn[k,-1] = self.boundary_conditions[2*k+3](x)

          self.logger.debug(
            'ion species {k:02d} BC residual Fn[{k:d},0]  = {:> 8.4g}'.format(
              Fn[k,0],k=k))
          self.logger.debug(
            'ion species {k:02d} BC residual Fn[{k:d},-1]  = {:> 8.4g}'.format(
              Fn[k,-1],k=k))

        return Fn

    # non-linear system, "controlled volume" method
    # Selbherr, S. Analysis and Simulation of Semiconductor Devices, Spriger 1984
    def G(self, x):
        """Non-linear system

        Discretization of Poisson-Nernst-Planck system with M ion species.
        Implements "controlled volume" method as found in

          Selbherr, Analysis and Simulation of Semiconductor Devices, Spriger 1984

        Parameters
        ----------
        x : ((M+1)*Ni,) ndarray
            system variables. 1D array of (M+1)*Ni values, wher M is number of
            ion sepcies, Ni number of spatial discretization points. First Ni
            entries are expected to contain potential, following M*Ni points
            contain ion concentrations.
        Returns
        --------
        residual: ((M+1)*Ni,) ndarray
        """
        # reduced Poisson equation: d2udx2 = rho
        Fu = self.potential_pde(x)
        Fn = self.concentration_pde(x)

        # Apply constraints if set (not implemented properly, do not use):
        if len(self.g) > 0:
            Flam = np.array([g(x) for g in self.g])
            F = np.concatenate([Fu,Fn.flatten(),Flam])
        else:
            F = np.concatenate([Fu,Fn.flatten()])

        return F

    def I(self): # ionic strength
        """Compute the system's ionic strength from charges and concentrations.

        Returns
        -------
        I : float
            ionic strength ( 1/2 * sum(z_i^2*c_i) )
            [concentration unit, i.e. mol m^-3]
        """
        return 0.5*np.sum( np.square(self.z) * self.c )

    def lambda_D(self):
        """Compute the system's Debye length.

        Returns
        -------
        lambda_D : float
            Debye length, sqrt( epsR*eps*R*T/(2*F^2*I) ) [length unit, i.e. m]
        """
        return np.sqrt(
            self.relative_permittivity*self.vacuum_permittivity*self.R*self.T/(
                2.0*self.F**2*self.I() ) )

    # default 0.1 mM (i.e. mol/m^3) NaCl aqueous solution
    def __init__(self,
        c = np.array([0.1,0.1]),
        z = np.array([1,-1]),
        L = 100e-9, # 100 nm
        lambda_S=0, # Stern layer (compact layer) thickness
        x0 = 0, # zero position
        T = 298.15,
        delta_u = 0.05, # potential difference [V]
        relative_permittivity = 79,
        vacuum_permittivity   = sc.epsilon_0,
        R = sc.value('molar gas constant'),
        F = sc.value('Faraday constant'),
        N = 200, # number of grid segments, number of grid points Ni = N + 1
        e = 1e-10, # absolute tolerance, TODO: switch to standaradized measure
        maxit = 20 ): # maximum number of Newton iterations
        """Initializes a 1D Poisson-Nernst-Planck system description.

        Expects quantities in SI units per default.

        Parameters
        ----------
        c : (M,) ndarray, optional
            bulk concentrations of each ionic species [mol/m^3]
            (default: [ 0.1, 0.1 ])
        z : (M,) ndarray, optional
            charge of each ionic species [1] (default: [ +1, -1 ])
        x0 : float, optional
            left hand side reference position (default: 0)
        L : float, optional
            1D domain size [m] (default: 100e-9)
        lambda_S: float, optional
            Stern layer thickness in case of Robin BC [m] (default: 0)
        T : float, optional
            temperature of the solution [K] (default: 298.15)
        delta_u : float, optional
            potential drop across 1D cell [V] (default: 0.05)
        relative_permittivity: float, optional
            relative permittivity of the ionic solution [1] (default: 79)
        vacuum_permittivity: float, optional
            vacuum permittivity [F m^-1] (default: 8.854187817620389e-12 )
        R : float, optional
            molar gas constant [J mol^-1 K^-1] (default: 8.3144598)
        F : float, optional
            Faraday constant [C mol^-1] (default: 96485.33289)
        N : int, optional
            number of discretization grid segments (default: 200)
        e : float, optional
            absolute tolerance for Newton solver convergence (default: 1e-10)
        maxit : int, optional
            maximum number of Newton iterations (default: 20)
        """

        self.logger = logging.getLogger(__name__)

        assert len(c) == len(z), "Provide concentration AND charge for ALL ion species!"

        # TODO: integrate with constructor initialization parameters above
        # default solver settings
        self.converged = False # solver's convergence flag
        self.N      = N     # discretization segments
        self.e      = e     # Newton solver default tolerance
        self.maxit  = maxit # Newton solver maximum iterations

        # default output settings
        # self.output = False   # let Newton solver output convergence plots...
        # self.outfreq = 1      # ...at every nth iteration
        self.label_width = 40 # charcater width of quantity labels in log

        # standard governing equations
        self.potential_pde      = self.poisson_pde
        self.concentration_pde  = self.nernst_planck_pde

        # empty BC
        self.boundary_conditions = []
        # empty constraints
        self.g = [] # list of constrain functions, not fully implemented / tested

        # system parameters
        self.M = len(c) # number of ion species

        self.c  = c # concentrations
        self.z  = z # number charges
        self.T  = T # temperature
        self.L  = L # 1d domain size
        self.lambda_S = lambda_S # Stern layer thickness
        self.x0 = x0 # reference position
        self.delta_u = delta_u # potential difference


        self.relative_permittivity = relative_permittivity
        self.vacuum_permittivity   = vacuum_permittivity
        # R = N_A * k_B
        # (universal gas constant = Avogadro constant * Boltzmann constant)
        self.R                     = R
        self.F                     = F

        self.f                     = F / (R*T) # for convenience

        # print all quantities to log
        for i, (c, z) in enumerate(zip(self.c,self.z)):
          self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            "ion species {:02d} concentration c".format(i), c, lwidth=self.label_width))
          self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            "ion species {:02d} number charge z".format(i), z, lwidth=self.label_width))

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'temperature T', self.T, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'domain size L', self.L, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'compact layer thickness lambda_S', self.lambda_S, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'reference position x0', self.x0, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'potential difference delta_u', self.delta_u, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'relative permittivity eps_R', self.relative_permittivity, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'vacuum permittivity eps_0', self.vacuum_permittivity, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'universal gas constant R', self.R, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'Faraday constant F', self.F, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'f = F / (RT)', self.f, lwidth=self.label_width))

        # scaled units for dimensionless formulation

        # length unit chosen as Debye length lambda
        self.l_unit = self.lambda_D()

        # concentration unit is ionic strength
        self.c_unit =  self.I()

        # no time unit for now, only steady state
        # self.t_unit = self.l_unit**2 / self.Dn # fixes Dn_scaled = 1

        self.u_unit = self.R * self.T / self.F # thermal voltage

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'spatial unit [l]', self.l_unit, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'concentration unit [c]', self.c_unit, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'potential unit [u]', self.u_unit, lwidth=self.label_width))

        # domain
        self.L_scaled = self.L / self.l_unit

        # compact layer
        self.lambda_S_scaled = self.lambda_S / self.l_unit

        # reference position
        self.x0_scaled = self.x0 / self.l_unit

        # bulk conectrations
        self.c_scaled = self.c / self.c_unit

        # potential difference
        self.delta_u_scaled = self.delta_u / self.u_unit


        # relaxation time
        # self.tau_scaled   = self.tau / self.t_unit

        # Dn_scaled    = Dn * t_unit / l_unit**2

        # print scaled quantities to log
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'reduced domain size L*', self.L_scaled, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
           'reduced compact layer thickness lambda_S*', self.lambda_S_scaled, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
           'reduced reference position x0*', self.x0_scaled, lwidth=self.label_width))

        for i, c_scaled in enumerate(self.c_scaled):
          self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            "ion species {:02d} reduced concentration c*".format(i),
            c_scaled, lwidth=self.label_width))

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'reduced potential delta_u*', self.delta_u_scaled, lwidth=self.label_width))


        # per default, no outer Helmholtz plane
        self.lhs_ohp = np.nan
        self.rhs_ohp = np.nan



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

from .base_qm_cluster_tool import BaseQMClusterTool
from ase import Atom


class QMClusteringTool(BaseQMClusterTool):
    """This class is responsible for carving and hydrogenating
    a qm cluster"""

    def __init__(self, mediator=None, double_bonded_atoms_list=[]):
        """This class is responsible for carving and hydrogenating
        a qm cluster

        Parameters
        ----------
        mediator : matscipy.calculators.mcfm.QMCluster
            class responsible for managing the QM clusters in the simulation
        double_bonded_atoms_list : list
            list of doubly bonded atoms, needed for double hydrogenation.
        """
        # Initialize the QMClusterObject with a mediator
        super(QMClusteringTool, self).__init__(mediator)

        self.double_bonded_atoms_list = double_bonded_atoms_list

    def create_buffer_region(self, atoms, qm_atoms_list, buffer_hops=10):
        """GIven a list of active QM atoms, returns a list containing buffer atoms indices

        Parameters
        ----------
        atoms : ase.Atoms
            whole structure
        qm_atoms_list : list of ints (atomic indexes)
            List of atoms in the inner QM region
        buffer_hops : int
            Expand the cluster by this many neighbour hops to create the buffer

        Returns
        -------
        list of ints (atomic indexes)
            buffer_list: List of atoms in the buffer region
        list of ints (atomic indexes)
            terminal_atoms: List of atoms from the buffer region that are on the verge of it
        list of ints (atomic indexes)
            cutoff_atoms_list: List of atoms that are not in the buffer but are bonded to the
            atoms in the buffer.
        """
        innerQM_region_set = set(qm_atoms_list)
        edge_neighbours = set(qm_atoms_list)
        terminal_atoms = []

        if len(qm_atoms_list) == len(atoms):
            return [], [], []

        for i in range(buffer_hops):
            new_neighbours = set()
            # For each atom in edge neighbours list, expand the list
            for index in edge_neighbours:
                new_neighbours |= set(self.find_neighbours(atoms, index)[0])
            # Remove atoms already in the qm list
            edge_neighbours = new_neighbours - innerQM_region_set

            # If the cluster is still growing, update the list of atoms at the edge
            if len(edge_neighbours) > 0:
                terminal_atoms = edge_neighbours
            # Make a union of the sets
            innerQM_region_set |= edge_neighbours

        # GO throught the loop one more time to find the last atoms not in the cluster
        new_neighbours = set()
        # For each atom in edge neighbours list, expand the list
        for index in edge_neighbours:
            new_neighbours |= set(self.find_neighbours(atoms, index)[0])
        # Remove atoms already in the qm list
        cutoff_atoms_set = new_neighbours - innerQM_region_set
        cutoff_atoms_list = list(cutoff_atoms_set)

        # Create buffer list
        innerQM_region_set -= set(qm_atoms_list)
        buffer_list = list(innerQM_region_set)
        terminal_atoms = list(terminal_atoms)
        return buffer_list, terminal_atoms, cutoff_atoms_list

    def carve_cluster(self, atoms, core_qm_list, buffer_hops=10):
        """Create a cluster with the list as core atoms, returns an ase.Atoms object

        Parameters
        ----------
        atoms : ase.Atoms
            whole structure
        core_qm_list : list of ints (atomic indexes)
            Indexes of atoms in the core QM region
        buffer_hops : int
            Expand the cluster by this many neighbour hops to create the buffer

        Returns
        -------
        ase.Atoms
            atoms object representing the QM cluster including
            the inner region and the buffer
        """

        # Define some lists
        total_supplementary_terminal_atoms = []

        # Buffer atoms - the buffer region
        # terminal_atoms_list - last atoms in the buffer region, bonded to the atoms not in the buffer
        # cutoff_atoms_list - atoms to be changed into hydrogen (cut-off from the buffer)
        # total_supplementary_terminal_atoms - atoms added to the buffer to make clusters more physical

        self.print_message("Creating buffer list", 1)
        buffer_list, terminal_atoms_list, cutoff_atoms_list =\
            self.create_buffer_region(atoms, core_qm_list, buffer_hops=buffer_hops)

        # If a spcial atoms was to be cut-off, add all of them to the cluster
        specialFlag = True
        while (specialFlag):
            specialFlag = False
            for specialMolecule in self.mediator.special_atoms_list:
                # Only operate on the special molecule if a part of it is inside the buffer
                if not any([specialAtom in buffer_list for specialAtom in specialMolecule]):
                    continue

                for specialAtomIndex in specialMolecule:
                    if (specialAtomIndex in cutoff_atoms_list):
                        self.include_special_atom(specialAtomIndex,
                                                  atoms,
                                                  buffer_list,
                                                  terminal_atoms_list,
                                                  cutoff_atoms_list)
                        # If at least one atom was added, cotinue to the next loop
                        specialFlag = True

        # Complete aromatic rings, repeat the process untill no more atoms are added
        iterMax = 10
        for i in range(iterMax):
            completeFlag = self.complete_aromatic_rings(
                atoms, buffer_list, terminal_atoms_list, cutoff_atoms_list,
                total_supplementary_terminal_atoms)
            if (not completeFlag):
                break

        # Create joint list. Buffer list is the original buffer while the
        # supplementary list is composed of new additions
        self.print_message("Creating joint listt", 10)
        if len(buffer_list) > 0:
            total_list = core_qm_list + buffer_list + total_supplementary_terminal_atoms
        else:
            total_list = core_qm_list

        # Add missing hydrogens (all routines operate only on heavy atoms)
        total_set = set(total_list)
        self.hydrogenate_cluster(atoms, total_set)
        total_list = list(total_set)

        self.print_message("finished adding atoms", 10)
        self.print_message("Buffer complete, creating cluster from mark", 1)

        atomic_cluster = self.create_cluster_from_marks(atoms, total_list)
        atomic_cluster.info["no_quantum_atoms"] = len(atomic_cluster)

        # Add properties for core region and buffer
        atomic_cluster.arrays["cluster_mark"] = np.zeros(len(atomic_cluster), dtype=int)
        atomic_cluster.arrays["cluster_mark"] += 5

        for i in range(len(atomic_cluster)):
            if atomic_cluster.arrays["orig_index"][i] in total_supplementary_terminal_atoms:
                atomic_cluster.arrays["cluster_mark"][i] = 4

            elif atomic_cluster.arrays["orig_index"][i] in terminal_atoms_list:
                atomic_cluster.arrays["cluster_mark"][i] = 3

            elif atomic_cluster.arrays["orig_index"][i] in buffer_list:
                atomic_cluster.arrays["cluster_mark"][i] = 2

            elif atomic_cluster.arrays["orig_index"][i] in core_qm_list:
                atomic_cluster.arrays["cluster_mark"][i] = 1

        # Change the cut-off atoms into hydrogens
        self.print_message("Change cutoff atoms into hydrogens", 1)
        if (len(cutoff_atoms_list) > 0) and(len(terminal_atoms_list) > 0):
            terminal_atoms_list = list(set(terminal_atoms_list))
            self.hydrogenate_dangling_bonds(terminal_atoms_list, cutoff_atoms_list, atomic_cluster, atoms)

        self.print_message("Center the atomic_cluster and remove PBC's", 1)
        # Center the cluster and remove PBC's
        # atomic_cluster.positions += np.array([0, 0, 20])
        atomic_cluster.wrap()
        atomic_cluster.center(vacuum=30)
        atomic_cluster.pbc = np.array([0, 0, 0], dtype=bool)

        self.print_message("Fished!", 1)

        return atomic_cluster

    def hydrogenate_dangling_bonds(self, terminal_atoms_list, cutoff_atoms_list, atomic_cluster, atoms):
        """Change atoms that were cut-off into hydrogens

        Parameters
        ----------
        terminal_atoms_list : list of ints (atomic indexes)
            last atoms in the buffer region, bonded to the atoms not in the buffer
        cutoff_atoms_list : list of ints (atomic indexes)
            atoms to be changed into hydrogen, first atoms not in the buffer
        atomic_cluster : ase.Atoms
            QM region structure (with core and buffer atoms)
        atoms : ase.Atoms
            whole structure
        """

        pos = atoms.get_positions()
        # Change cutoff list into a numpy array
        cutoff_atoms_list = np.asarray(cutoff_atoms_list)
        for tAI in terminal_atoms_list:
            # Check if any of the cut off atoms are neighbours of the terminal atom
            cutoff_neighs = [item for item in self.mediator.neighbour_list.get_neighbours(
                tAI) if (item in cutoff_atoms_list)]
            # Iterate over all cut-off atoms that are neighburs of tAI to
            # Effectively loop over all cut bonds
            for cAI in cutoff_neighs:
                if ((cAI in self.double_bonded_atoms_list) and (tAI in self.double_bonded_atoms_list)):
                    self.replace_double_bond(tAI, cAI, atomic_cluster, atoms, pos)
                else:
                    self.replace_single_bond(tAI, cAI, atomic_cluster, atoms, pos)

    def replace_single_bond(self,
                            terminal_atom_index,
                            cutoff_atom_index,
                            atomic_cluster,
                            atoms,
                            atomic_positions):
        """Replace a cut-off atom with a single hydrogen

        Parameters
        ----------
        terminal_atoms_list : list of ints (atomic indexes)
            last atoms in the buffer region, bonded to the atoms not in the buffer
        cutoff_atoms_list : list of ints (atomic indexes)
            atoms to be changed into hydrogen, first atoms not in the buffer
        atomic_cluster : ase.Atoms
            QM region structure (with core and buffer atoms)
        atoms : ase.Atoms
            whole structure
        atomic_positions : np.array
            Positions of atoms in the whole structure (copy of the atoms.positions)
        """
        vector = atomic_positions[cutoff_atom_index] - atomic_positions[terminal_atom_index]
        # Make the bond approximately 1 angstrom
        vector /= np.linalg.norm(vector)
        vector *= 1
        # Add a hydrogen instead of the cutoff atom
        pos = atomic_positions[terminal_atom_index] + vector
        cutoff_hydro = Atom(symbol=1, position=pos, charge=0.1)
        atomic_cluster.append(cutoff_hydro)
        atomic_cluster.arrays["orig_index"][len(atomic_cluster) - 1] = len(atoms) + 1
        atomic_cluster.arrays["cluster_mark"][len(atomic_cluster) - 1] = 6

    def replace_double_bond(self,
                            terminal_atom_index,
                            cutoff_atom_index,
                            atomic_cluster,
                            atoms,
                            atomic_positions):
        """Replace a cut-off atom with two hydrogens

        Parameters
        ----------
        terminal_atoms_list : list of ints (atomic indexes)
            last atoms in the buffer region, bonded to the atoms not in the buffer
        cutoff_atoms_list : list of ints (atomic indexes)
            atoms to be changed into hydrogen, first atoms not in the buffer
        atomic_cluster : ase.Atoms
            QM region structure (with core and buffer atoms)
        atoms : ase.Atoms
            whole structure
        atomic_positions : np.array
            Positions of atoms in the whole structure (copy of the atoms.positions)
        """
        # Find a vector to from the terminal atom to the cutoff atom
        vector = atomic_positions[cutoff_atom_index] - atomic_positions[terminal_atom_index]

        # ------ Find the displacement between two hydrogens
        # Find two closest neighbours of the cut-off atom
        neighbours = np.asarray(self.find_neighbours(atoms, terminal_atom_index)[0])
        dispVectors = atomic_positions[neighbours] - atomic_positions[terminal_atom_index]
        distances = np.sum(np.square(dispVectors), axis=1)
        closeNeighbours = np.argsort(distances)
        closeNeighbours = neighbours[closeNeighbours][: 2]

        # Find the vectors to those two atoms
        a1 = atomic_positions[terminal_atom_index] - atomic_positions[closeNeighbours[0]]
        a2 = atomic_positions[terminal_atom_index] - atomic_positions[closeNeighbours[1]]

        # Find the cross product of a1 and a2 thus finding a vector perpendicular to
        # the plane they define
        aPerp = np.cross(a1, a2)
        aPerp /= np.linalg.norm(aPerp)
        aPerp *= 2

        # Create two vectors, the initial displacement +/- the perpendicular vector
        vector1 = vector + aPerp
        vector2 = vector - aPerp

        # Make the bonds approximately 1 angstrom
        vector1 /= np.linalg.norm(vector1)
        vector2 /= np.linalg.norm(vector2)

        # Add a hydrogen instead of the cutoff atom
        pos = atomic_positions[terminal_atom_index] + vector1
        cutoff_hydro = Atom(symbol=1, position=pos, charge=0.1)
        atomic_cluster.append(cutoff_hydro)
        atomic_cluster.arrays["orig_index"][len(atomic_cluster) - 1] = len(atoms) + 1
        atomic_cluster.arrays["cluster_mark"][len(atomic_cluster) - 1] = 6

        pos = atomic_positions[terminal_atom_index] + vector2
        cutoff_hydro = Atom(symbol=1, position=pos, charge=0.1)
        atomic_cluster.append(cutoff_hydro)
        atomic_cluster.arrays["orig_index"][len(atomic_cluster) - 1] = len(atoms) + 1
        atomic_cluster.arrays["cluster_mark"][len(atomic_cluster) - 1] = 6

    def include_special_atom(self,
                             specialAtomIndex,
                             atoms,
                             buffer_list,
                             terminal_atoms_list,
                             cutoff_atoms_list):
        """Add a special atom to the buffer and update indexes.
        In case a group of special atoms are specified (special molecule),
        If one of these atoms is in the buffer regios, the rest are also added to it.

        Parameters
        ----------
        specialAtomIndex : int (atomic index)
            Add a specified atoms to the buffer
        atoms : ase.Atoms
            whole structure
        buffer_list : list of ints (atomic indexes)
            List of atoms in the buffer region
        terminal_atoms_list : list of ints (atomic indexes)
            last atoms in the buffer region, bonded to the atoms not in the buffer
        cutoff_atoms_list : list of ints (atomic indexes)
            atoms to be changed into hydrogen, first atoms not in the buffer
        """

        buffer_list.append(specialAtomIndex)
        terminal_atoms_list.append(specialAtomIndex)
        cutoff_atoms_list.remove(specialAtomIndex)

        # ------ Add new cutoff atoms
        specialAtomNeighbours = self.find_neighbours(atoms, specialAtomIndex)[0]
        for neighIndex in specialAtomNeighbours:
            if (neighIndex not in buffer_list) and (neighIndex not in cutoff_atoms_list):
                cutoff_atoms_list.append(neighIndex)

    def complete_aromatic_rings(self, atoms, buffer_list,
                                terminal_atoms_list, cutoff_atoms_list,
                                total_supplementary_terminal_atoms):
        """Check if a terminal atom is not connected ot two atoms at once
        If it is, add it. This deals with aromatic ring structures

        Parameters
        ----------
        atoms : ase.Atoms
            whole structure
        buffer_list : list of ints (atomic indexes)
            List of atoms in the buffer region
        terminal_atoms_list : list of ints (atomic indexes)
            last atoms in the buffer region, bonded to the atoms not in the buffer
        cutoff_atoms_list : list of ints (atomic indexes)
            atoms to be changed into hydrogen, first atoms not in the buffer
        total_supplementary_terminal_atoms : NONE
            atoms added to the buffer to make clusters more physical.
            Example: last atom in the aromnatic ring

        Returns
        -------
        bool
            Return True if any atoms were added
        """

        supplementary_terminal_atoms = []

        # Buffer atoms - the buffer region
        # terminal_atoms_list - last atoms in the buffer region, bonded to the atoms not in the buffer
        # cutoff_atoms_list - atoms to be changed into hydrogen (cut-off from the buffer)
        # supplementary_terminal_atoms - atoms added to the buffer to make clusters more physical

        self.print_message("Completing destroyed rings", 1)
        for index, cI in enumerate(cutoff_atoms_list):
            msg = "Working on atom {0} with number {1}".format(cI, atoms.numbers[cI])
            self.print_message(msg, 100)

            # Check if a cutoff atom has more than 1 neighbour in terminal atoms list
            neighs = self.find_neighbours(atoms, cI)[0]
            cutoff_atom_neighs = [item for item in neighs if item in (terminal_atoms_list + buffer_list)]
            # If two or more, add it
            if (len(cutoff_atom_neighs) >= 2):
                supplementary_terminal_atoms.append(cI)
                self.print_message("Adding {0} to supplementary index".format(cI), 10)
        self.print_message("Finished adding atoms.", 10)

        # Return False if no atoms were added
        if (len(supplementary_terminal_atoms) == 0):
            return False

        # Keep track of all the additions to the supplementary atoms
        total_supplementary_terminal_atoms += supplementary_terminal_atoms
        terminal_atoms_list += total_supplementary_terminal_atoms
        if (self.verbose >= 10):
            print("Added", len(supplementary_terminal_atoms), "Atoms!")
        if (self.verbose >= 100):
            print("Added list:", supplementary_terminal_atoms)

        # Find new cutoff atoms
        if (self.verbose >= 10):
            print("Finding new cutoff atoms")

        cutoff_atoms_list[:] = []
        outer_qm_list = buffer_list + total_supplementary_terminal_atoms
        for eqI in terminal_atoms_list:
            new_cutoff_atoms = [item for item in self.find_neighbours(atoms, eqI)[0] if
                                (item not in outer_qm_list)]
            cutoff_atoms_list += new_cutoff_atoms

        # Get rid of duplicates
        cutoff_atoms_list[:] = list(set(cutoff_atoms_list))
        terminal_atoms_list[:] = list(set(terminal_atoms_list))

        # Return True if any atoms were added:
        return True

    def create_cluster_from_marks(self, atoms, select_list):
        """Return an ase.Atoms object containing selected atoms from
        a larger structure

        Parameters
        ----------
        atoms : ase.Atoms
            whole structure
        select_list : list of ints (atomic indexes)
            List of atoms to include in the new structure

        Returns
        -------
        ase.Atoms
            Structure composed of selected atoms
        """

        if (len(select_list) > len(atoms)):
            select_list = np.unique(select_list)

        pbc = atoms.get_pbc()
        cell = atoms.get_cell()
        cluster = atoms.__class__(cell=cell, pbc=pbc)

        cluster.arrays = {}
        # for name, a in atoms.arrays.items():
        #     cluster.arrays[name] = a[select_list].copy()

        for name in ["numbers", "positions"]:
            cluster.arrays[name] = atoms.arrays[name][select_list].copy()

        cluster.arrays["orig_index"] = np.asarray(select_list, dtype=int)

        return cluster

    def print_message(self, message, limit=100):
        """Print a message if the calculators verbosity level is above the
        given threshold

        Parameters
        ----------
        message : str
            The message to be printed
        limit : int
            the verbosity threshold for this mesage
        """
        if (self.verbose >= limit):
            print(message)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from __future__ import print_function

import itertools
import warnings

import numpy as np
from numpy.linalg import inv, norm
try:
    import scipy.stats as scipy_stats
except:
    scipy_stats = None

import ase.units as units
from ase.atoms import Atoms

###

# The indices of the full stiffness matrix of (orthorhombic) interest
Voigt_notation = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]

def full_3x3_to_Voigt_6_index(i, j):
    if i == j:
        return i
    return 6-i-j

###

def Voigt_6_to_full_3x3_strain(strain_vector):
    """
    Form a 3x3 strain matrix from a 6 component vector in Voigt notation
    """
    e1, e2, e3, e4, e5, e6 = np.transpose(strain_vector)
    return np.transpose([[1.0+e1, 0.5*e6, 0.5*e5],
                         [0.5*e6, 1.0+e2, 0.5*e4],
                         [0.5*e5, 0.5*e4, 1.0+e3]])


def Voigt_6_to_full_3x3_stress(stress_vector):
    """
    Form a 3x3 stress matrix from a 6 component vector in Voigt notation
    """
    s1, s2, s3, s4, s5, s6 = np.transpose(stress_vector)
    return np.transpose([[s1, s6, s5],
                         [s6, s2, s4],
                         [s5, s4, s3]])


def full_3x3_to_Voigt_6_strain(strain_matrix):
    """
    Form a 6 component strain vector in Voigt notation from a 3x3 matrix
    """
    strain_matrix = np.asarray(strain_matrix)
    return np.transpose([strain_matrix[...,0,0] - 1.0,
                         strain_matrix[...,1,1] - 1.0,
                         strain_matrix[...,2,2] - 1.0,
                         strain_matrix[...,1,2]+strain_matrix[...,2,1],
                         strain_matrix[...,0,2]+strain_matrix[...,2,0],
                         strain_matrix[...,0,1]+strain_matrix[...,1,0]])


def full_3x3_to_Voigt_6_stress(stress_matrix):
    """
    Form a 6 component stress vector in Voigt notation from a 3x3 matrix
    """
    stress_matrix = np.asarray(stress_matrix)
    return np.transpose([stress_matrix[...,0,0],
                         stress_matrix[...,1,1],
                         stress_matrix[...,2,2],
                         (stress_matrix[...,1,2]+stress_matrix[...,1,2])/2,
                         (stress_matrix[...,0,2]+stress_matrix[...,0,2])/2,
                         (stress_matrix[...,0,1]+stress_matrix[...,0,1])/2])


def Voigt_6x6_to_full_3x3x3x3(C):
    """
    Convert from the Voigt representation of the stiffness matrix to the full
    3x3x3x3 representation.

    Parameters
    ----------
    C : array_like
        6x6 stiffness matrix (Voigt notation).
    
    Returns
    -------
    C : array_like
        3x3x3x3 stiffness matrix.
    """
    
    C = np.asarray(C)
    C_out = np.zeros((3,3,3,3), dtype=float)
    for i, j, k, l in itertools.product(range(3), range(3), range(3), range(3)):
        Voigt_i = full_3x3_to_Voigt_6_index(i, j)
        Voigt_j = full_3x3_to_Voigt_6_index(k, l)
        C_out[i, j, k, l] = C[Voigt_i, Voigt_j]
    return C_out


def full_3x3x3x3_to_Voigt_6x6(C):
    """
    Convert from the full 3x3x3x3 representation of the stiffness matrix
    to the representation in Voigt notation. Checks symmetry in that process.
    """

    tol = 1e-3

    C = np.asarray(C)
    Voigt = np.zeros((6,6))
    for i in range(6):
        for j in range(6):
            k, l = Voigt_notation[i]
            m, n = Voigt_notation[j]
            Voigt[i,j] = C[k,l,m,n]

            #print '---'
            #print k,l,m,n, C[k,l,m,n]
            #print m,n,k,l, C[m,n,k,l]
            #print l,k,m,n, C[l,k,m,n]
            #print k,l,n,m, C[k,l,n,m]
            #print m,n,l,k, C[m,n,l,k]
            #print n,m,k,l, C[n,m,k,l]
            #print l,k,n,m, C[l,k,n,m]
            #print n,m,l,k, C[n,m,l,k]
            #print '---'

            # Check symmetries
            assert abs(Voigt[i,j]-C[m,n,k,l]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], m, n, k, l, C[m,n,k,l])
            assert abs(Voigt[i,j]-C[l,k,m,n]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], k, l, m, n, C[l,k,m,n])
            assert abs(Voigt[i,j]-C[k,l,n,m]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], k, l, n, m, C[k,l,n,m])
            assert abs(Voigt[i,j]-C[m,n,l,k]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], m, n, l, k, C[m,n,l,k])
            assert abs(Voigt[i,j]-C[n,m,k,l]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], n, m, k, l, C[n,m,k,l])
            assert abs(Voigt[i,j]-C[l,k,n,m]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], l, k, n, m, C[l,k,n,m])
            assert abs(Voigt[i,j]-C[n,m,l,k]) < tol, \
                'Voigt[{},{}] = {}, C[{},{},{},{}] = {}' \
                .format(i, j, Voigt[i,j], n, m, l, k, C[n,m,l,k])

    return Voigt


def Voigt_6x6_to_cubic(C):
    """
    Convert the Voigt 6x6 representation into the cubic elastic constants
    C11, C12 and C44.
    """

    tol = 1e-6

    C_check = np.zeros_like(C)
    C_check[np.diag_indices_from(C_check)] = C[np.diag_indices_from(C)]
    C_check[0:3,0:3] = C[0:3,0:3]
    if np.any(np.abs(C-C_check) > tol):
        raise ValueError('"C" does not have cubic symmetry.')

    C11s = np.array([C[0,0], C[1,1], C[2,2]])
    C12s = np.array([C[1,2], C[0,2], C[0,1]])
    C44s = np.array([C[3,3], C[4,4], C[5,5]])

    C11 = np.mean(C11s)
    C12 = np.mean(C12s)
    C44 = np.mean(C44s)

    if np.any(np.abs(C11-C11s) > tol) or np.any(np.abs(C12-C12s) > tol) or \
            np.any(np.abs(C44-C44s) > tol):
        raise ValueError('"C" does not have cubic symmetry.')

    return np.array([C11, C12, C44])


def cubic_to_Voigt_6x6(C11, C12, C44):
    return np.array([[C11,C12,C12,  0,  0,  0],
                     [C12,C11,C12,  0,  0,  0],
                     [C12,C12,C11,  0,  0,  0],
                     [  0,  0,  0,C44,  0,  0],
                     [  0,  0,  0,  0,C44,  0],
                     [  0,  0,  0,  0,  0,C44]])

def _invariants(s, syy=None, szz=None, syz=None, sxz=None, sxy=None,
               full_3x3_to_Voigt_6=full_3x3_to_Voigt_6_stress):
    """
    Receives a list of stress tensors and returns the three tensor invariants.
    """
    if syy is None:
        s = np.asarray(s)
        if s.shape == (6,):
            s = s.reshape(1,-1)
        elif s.shape == (3,3):
            s = full_3x3_to_Voigt_6(s)
        elif s.shape[-1] == 3 and s.shape[-2] == 3:
            s = full_3x3_to_Voigt_6(s)
    else:
        s = np.transpose([np.transpose(s),
                          np.transpose(syy),
                          np.transpose(szz),
                          np.transpose(syz),
                          np.transpose(sxz),
                          np.transpose(sxy)])
    I1 = s[...,0]+s[...,1]+s[...,2]
    I2 = s[...,0]*s[...,1]+s[...,1]*s[...,2]+s[...,2]*s[...,0]-s[...,3]**2- \
        s[...,4]**2-s[...,5]**2
    I3 = s[...,0]*s[...,1]*s[...,2]+2*s[...,3]*s[...,4]*s[...,5]- \
        s[...,3]**2*s[...,2]-s[...,4]**2*s[...,0]-s[...,5]**2*s[...,1]

    return I1, I2, I3

def invariants(s, syy=None, szz=None, syz=None, sxz=None, sxy=None,
               full_3x3_to_Voigt_6=full_3x3_to_Voigt_6_stress):
    I1, I2, I3 = _invariants(s, syy=syy, szz=szz, syz=syz, sxz=sxz, sxy=sxy,
                             full_3x3_to_Voigt_6=full_3x3_to_Voigt_6)

    J2 = I1**2/3-I2
    J3 = 2*I1**3/27-I1*I2/3+I3

    # Return hydrostatic pressure, octahedral shear stress and J3
    return -I1/3, np.sqrt(2*J2/3), J3

###

def rotate_cubic_elastic_constants(C11, C12, C44, A, tol=1e-6):
    """
    Return rotated elastic moduli for a cubic crystal given the elastic 
    constant in standard C11, C12, C44 notation.

    Parameters
    ----------
    C11, C12, C44 : float
        Cubic elastic moduli.
    A : array_like
        3x3 rotation matrix.

    Returns
    -------
    C : array
        6x6 matrix of rotated elastic constants (Voigt notation).
    """

    A = np.asarray(A)

    # Is this a rotation matrix?
    if np.sometrue(np.abs(np.dot(np.array(A), np.transpose(np.array(A))) - 
                          np.eye(3, dtype=float)) > tol):
        raise RuntimeError('Matrix *A* does not describe a rotation.')

    # Invariant elastic constants
    la = C12
    mu = C44
    al = C11 - la - 2*mu

    # Construct rotated C in Voigt notation
    C = [ ]
    for i, j in Voigt_notation:
        for k, l in Voigt_notation:
            h = 0.0
            if i == j and k == l:
                h += la
            if i == k and j == l:
                h += mu
            if i == l and j == k:
                h += mu
            h += al*np.sum(A[i,:]*A[j,:]*A[k,:]*A[l,:])
            C += [ h ]

    C = np.array(C)
    C.shape = (6, 6)
    return C

###

def rotate_elastic_constants(C, A, tol=1e-6):
    """
    Return rotated elastic moduli for a general crystal given the elastic 
    constant in Voigt notation.

    Parameters
    ----------
    C : array_like
        6x6 matrix of elastic constants (Voigt notation).
    A : array_like
        3x3 rotation matrix.

    Returns
    -------
    C : array
        6x6 matrix of rotated elastic constants (Voigt notation).
    """

    A = np.asarray(A)

    # Is this a rotation matrix?
    if np.sometrue(np.abs(np.dot(np.array(A), np.transpose(np.array(A))) - 
                          np.eye(3, dtype=float)) > tol):
        raise RuntimeError('Matrix *A* does not describe a rotation.')

    # Rotate
    return full_3x3x3x3_to_Voigt_6x6(np.einsum('ia,jb,kc,ld,abcd->ijkl',
                                               A, A, A, A,
                                               Voigt_6x6_to_full_3x3x3x3(C)))

###

class CubicElasticModuli:
    tol = 1e-6

    def __init__(self, C11, C12, C44):
        """
        Initialize a cubic system with elastic constants C11, C12, C44
        """

        warnings.warn('CubicElasticModuli is deprecated. Use '
                      'rotate_elastic_constants function instead.')

        # la, mu, al are the three invariant elastic constants
        self.la = C12
        self.mu = C44
        self.al = C11 - self.la - 2*self.mu

        A = np.eye(3, dtype=float)

        # Compute initial stiffness matrix
        self.rotate(A)


    def rotate(self, A):
        """
        Compute the rotated stiffness matrix
        """

        A = np.asarray(A)

        # Is this a rotation matrix?
        if np.sometrue(np.abs(np.dot(np.array(A), np.transpose(np.array(A))) - 
                              np.eye(3, dtype=float)) > self.tol):
            raise RuntimeError('Matrix *A* does not describe a rotation.')

        C = [ ]
        for i, j in Voigt_notation:
            for k, l in Voigt_notation:
                h = 0.0
                if i == j and k == l:
                    h += self.la
                if i == k and j == l:
                    h += self.mu
                if i == l and j == k:
                    h += self.mu
                h += self.al*np.sum(A[i,:]*A[j,:]*A[k,:]*A[l,:])
                C += [ h ]

        self.C = np.asarray(C)
        self.C.shape = (6, 6)
        return self.C


    def _rotate_explicit(self, A):
        """
        Compute the rotated stiffness matrix by applying the rotation to the
        full stiffness matrix. This function is for debugging purposes only.
        """

        A = np.asarray(A)

        # Is this a rotation matrix?
        if np.sometrue(np.abs(np.dot(np.array(A), np.transpose(np.array(A))) - 
                              np.eye(3, dtype=float) ) > self.tol):
            raise RuntimeError('Matrix *A* does not describe a rotation.')

        C = np.zeros((3, 3, 3, 3), dtype=float)

        # Construct unrotated stiffness matrix
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    for m in range(3):
                        h = 0.0
                        if i == j and k == m:
                            h += self.la
                        if i == k and j == m:
                            h += self.mu
                        if i == m and j == k:
                            h += self.mu
                        if i == j and j == k and k == m:
                            h += self.al
                        C[i, j, k, m] = h

        # Rotate
        C = np.einsum('ia,jb,kc,ld,abcd->ijkl', A, A, A, A, C)

        self.C = full_3x3x3x3_to_Voigt_6x6(C)
        return self.C


    def stiffness(self):
        """
        Return the elastic constants
        """

        return self.C


    def compliance(self):
        """
        Return the compliance coefficients
        """

        return inv(self.C)

###

def measure_triclinic_elastic_constants(a, delta=0.001, optimizer=None, 
                                        logfile=None, **kwargs):
    """
    Brute-force measurement of elastic constants for a triclinic (general)
    unit cell.

    Parameters
    ----------
    a : ase.Atoms
        Atomic configuration.
    optimizer : ase.optimizer.*
        Optimizer to use for atomic position. Does not optimize atomic
        position if set to None.
    delta : float
        Strain increment for analytical derivatives of stresses.

    Returns
    -------
    C : array_like
        6x6 matrix of the elastic constants in Voigt notation.
    """

    if optimizer is not None:
        optimizer(a, logfile=logfile).run(**kwargs)

    r0 = a.positions.copy()

    cell = a.cell.copy()
    volume = a.get_volume()

    C = np.zeros((3,3,3,3), dtype=float)
    for i in range(3):
        for j in range(3):
            a.set_cell(cell, scale_atoms=True)
            a.set_positions(r0)
        
            D = np.eye(3)
            D[i, j] += 0.5*delta
            D[j, i] += 0.5*delta
            a.set_cell(np.dot(D, cell.T).T, scale_atoms=True)
            if optimizer is not None:
                optimizer(a, logfile=logfile).run(**kwargs)
            sp = Voigt_6_to_full_3x3_stress(a.get_stress()*a.get_volume())

            D = np.eye(3)
            D[i, j] -= 0.5*delta
            D[j, i] -= 0.5*delta
            a.set_cell(np.dot(D, cell.T).T, scale_atoms=True)
            if optimizer is not None:
                optimizer(a, logfile=logfile).run(**kwargs)
            sm = Voigt_6_to_full_3x3_stress(a.get_stress()*a.get_volume())

            C[:,:,i,j] = (sp-sm)/(2*delta*volume)

    a.set_cell(cell, scale_atoms=True)
    a.set_positions(r0)

    return full_3x3x3x3_to_Voigt_6x6(C)



Cij_symmetry = {
   'cubic':           np.array([[1, 7, 7, 0, 0, 0],
                                [7, 1, 7, 0, 0, 0],
                                [7, 7, 1, 0, 0, 0],
                                [0, 0, 0, 4, 0, 0],
                                [0, 0, 0, 0, 4, 0],
                                [0, 0, 0, 0, 0, 4]]),

   'trigonal_high':   np.array([[1, 7, 8, 9, 10, 0],
                                [7, 1, 8, 0,-9, 0],
                                [8, 8, 3, 0, 0, 0],
                                [9, -9, 0, 4, 0, 0],
                                [10, 0, 0, 0, 4, 0],
                                [0, 0, 0, 0, 0, 6]]),

   'trigonal_low':    np.array([[1,  7,  8,  9,  10,  0 ],
                                [7,  1,  8, -9, -10,  0 ],
                                [8,  8,  3,  0,   0,  0 ],
                                [9, -9,  0,  4,   0, -10],
                                [10,-10, 0,  0,   4,  9 ],
                                [0,  0,  0, -10 , 9,  6 ]]),

   'tetragonal_high': np.array([[1, 7, 8, 0, 0, 0],
                                [7, 1, 8, 0, 0, 0],
                                [8, 8, 3, 0, 0, 0],
                                [0, 0, 0, 4, 0, 0],
                                [0, 0, 0, 0, 4, 0],
                                [0, 0, 0, 0, 0, 6]]),

   'tetragonal_low':  np.array([[1, 7, 8, 0, 0, 11],
                                [7, 1, 8, 0, 0, -11],
                                [8, 8, 3, 0, 0, 0],
                                [0, 0, 0, 4, 0, 0],
                                [0, 0, 0, 0, 4, 0],
                                [11, -11, 0, 0, 0, 6]]),

   'orthorhombic':    np.array([[ 1,  7,  8,  0,  0,  0],
                                [ 7,  2, 12,  0,  0,  0],
                                [ 8, 12,  3,  0,  0,  0],
                                [ 0,  0,  0,  4,  0,  0],
                                [ 0,  0,  0,  0,  5,  0],
                                [ 0,  0,  0,  0,  0,  6]]),

   'monoclinic':      np.array([[ 1,  7,  8,  0,  10,  0],
                                [ 7,  2, 12,  0, 14,  0],
                                [ 8, 12,  3,  0, 17,  0],
                                [ 0,  0,  0,  4,  0,  20],
                                [10, 14, 17,  0,  5,  0],
                                [ 0,  0,  0, 20,  0,  6]]),
    
    'triclinic':       np.array([[ 1,  7,  8,  9,  10, 11],
                                 [ 7,  2, 12,  13, 14, 15],
                                 [ 8, 12,  3,  16, 17, 18],
                                 [ 9, 13, 16,  4,  19, 20],
                                 [10, 14, 17, 19,  5,  21],
                                 [11, 15, 18, 20,  21, 6 ]]),
   }

def _dec(pattern):
    # Decrease C_ij indices patterns by one.
    # Used to make strain_patterns definitions below more readable.
    return [(i-1, j-1) for (i,j) in pattern]

strain_patterns = {

   'cubic': [
      # strain pattern e1+e4, yields C11, C21, C31 and C44, then C12 is average of C21 and C31
      [ np.array([1,0,0,1,0,0]), _dec([(1,1), (2,1), (3,1), (4,4)])]
   ],

   'trigonal_high': [
      # strain pattern e3 yield C13, C23 and C33
      [ np.array([0,0,1,0,0,0]), _dec([(1,3), (2,3), (3,3)])],

      # strain pattern e1+e4 yields C11 C21 C31 and C44
      [ np.array([1,0,0,1,0,0]), _dec([(1,1), (2,1), (3,1), (4,4)])],

      # strain pattern e1 yields C11 C21 C31 C41 C51
      [ np.array([1,0,0,0,0,0]), _dec([(1,1), (2,1), (3,1), (4,1), (5,1)])],

      # strain pattern e3+e4
      [ np.array([0,0,1,1,0,0]), _dec([(3,3), (4,4)])]

   ],

   'trigonal_low': [
     # strain pattern e1, yields C11, C21, C31, C41, C51
     [ np.array([1,0,0,0,0,0]), _dec([(1,1), (2,1), (3,1), (4,1), (5,1)])],

     # strain pattern e3 + e4, yields C33, C44
     [ np.array([0,0,1,1,0,0]), _dec([(3,3), (4,4)]) ],

     [ np.array([0,0,0,0,0,1]), _dec([(6,6)]) ]
   ],

   'tetragonal': [
     # strain pattern e1+e4
     [ np.array([1,0,0,1,0,0]), _dec([(1,1), (2,1), (3,1), (6,1), (4,4)]) ],

     # strain pattern e3+e6
     [ np.array([0,0,1,0,0,1]), _dec([(3,3), (6,6)]) ]
   ],

   'orthorhombic': [
      # strain pattern e1+e4
      [ np.array([1,0,0,1,0,0]), _dec([(1,1), (2,1), (3,1), (4,4)]) ],

      # strain pattern e2+e5
      [ np.array([0,1,0,0,1,0]), _dec([(1,2), (2,2), (3,2), (5,5)]) ],

      # strain pattern e3+e6
      [ np.array([0,0,1,0,0,1]), _dec([(1,3), (2,3), (3,3), (6,6)]) ]
   ],

   'monoclinic': [
      # strain pattern e1+e4
      [ np.array([1,0,0,1,0,0]), _dec([(1,1), (2,1), (3,1), (4,4), (5,1), (6,4)]) ],

      # strain pattern e3+e6
      [ np.array([0,0,1,0,0,1]), _dec([(1,3), (2,3), (3,3), (5,3), (4,6), (6,6)]) ],

      # strain pattern e2
      [ np.array([0,1,0,0,0,0]), _dec([(1,2), (2,2), (3,2), (5,2)]) ],

      # strain pattern e5
      [ np.array([0,0,0,0,1,0]), _dec([(1,5), (2,5), (3,5), (5,5)]) ]
   ],

   'triclinic': [
      [ np.array([1,0,0,0,0,0]), _dec([(1,1), (2,1), (3,1), (4,1), (5,1), (6,1)])],
      [ np.array([0,1,0,0,0,0]), _dec([(1,2), (2,2), (3,2), (4,2), (5,2), (6,2)])],
      [ np.array([0,0,1,0,0,0]), _dec([(1,3), (2,3), (3,3), (4,3), (5,3), (6,3)])],
      [ np.array([0,0,0,1,0,0]), _dec([(1,4), (2,4), (3,4), (4,4), (5,4), (6,4)])],
      [ np.array([0,0,0,0,1,0]), _dec([(1,5), (2,5), (3,5), (4,5), (5,5), (6,5)])],
      [ np.array([0,0,0,0,0,1]), _dec([(1,6), (2,6), (3,6), (4,6), (5,6), (6,6)])],
   ]

   }

Cij_symmetry['hexagonal'] = Cij_symmetry['trigonal_high']
Cij_symmetry[None] = Cij_symmetry['triclinic']

strain_patterns['hexagonal'] = strain_patterns['trigonal_high']
strain_patterns['tetragonal_high'] = strain_patterns['tetragonal_low'] = strain_patterns['tetragonal']
strain_patterns[None] = strain_patterns['triclinic']


def generate_strained_configs(at0, symmetry='triclinic', N_steps=5, delta=1e-2):
    """
    Generate a sequence of strained configurations

    Parameters
    ----------
    a : ase.Atoms
        Bulk crystal configuration - both unit cell and atomic positions
        should be relaxed before calling this routine.
    symmetry : string
        Symmetry to use to determine which strain patterns are necessary.
        Default is 'triclininc', i.e. no symmetry.
    N_steps : int
        Number of atomic configurations to generate for each strain pattern.
        Default is 5. Absolute strain values range from -delta*N_steps/2
        to +delta*N_steps/2.
    delta : float
        Strain increment for analytical derivatives of stresses. Default 1e-2

    Returns
    -------
    Generator which yields a sequence of ase.Atoms instances corresponding
    to the minima set of strained conifugurations required to evaluate the
    full 6x6 C_ij matrix under the assumed symmetry.
    """

    if not symmetry in strain_patterns:
        raise ValueError('Unknown symmetry %s. Valid options are %s' % (symmetry, strain_patterns.keys()))

    for pindex, (pattern, fit_pairs) in enumerate(strain_patterns[symmetry]):
        for step in range(N_steps):
            strain = np.where(pattern == 1, delta*(step+1-(N_steps+1)/2.0), 0.0)
            at = at0.copy()
            if at0.get_calculator() is not None:
                at.set_calculator(at0.get_calculator())
            T = Voigt_6_to_full_3x3_strain(strain)
            at.set_cell(np.dot(T, at.cell.T).T, scale_atoms=False)
            at.positions[:] = np.dot(T, at.positions.T).T
            at.info['strain'] = T
            yield at


# Elastic constant calculation.

# Code adapted from elastics.py script, available from
# http://github.com/djw/elastic-constants
#
# Copyright (c) 2008, Dan Wilson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY DAN WILSON ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL DAN WILSON BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

def fit_elastic_constants(a, symmetry='triclinic', N_steps=5, delta=1e-2, optimizer=None,
                          verbose=True, graphics=False, logfile=None, **kwargs):
    """
    Compute elastic constants by linear regression of stress vs. strain

    Parameters
    ----------
    a : ase.Atoms or list of ase.Atoms
        Either a single atomic configuration or a list of strained
        configurations. If a single configuration is given, it is
        passed :func:`generate_strained_configs` along with `symmetry`, `N_steps`,
        and `delta` to generate the set of strained configurations.
    symmetry : string
        Symmetry to use to determine which strain patterns are necessary.
        Default is 'triclininc', i.e. no symmetry.
    N_steps : int
        Number of atomic configurations to generate for each strain pattern.
        Default is 5. Absolute strain values range from -delta*N_steps/2
        to +delta*N_steps/2.    
    delta : float
        Strain increment for analytical derivatives of stresses.
        Default is 1e-2.
    optimizer : ase.optimizer.*
        Optimizer to use for atomic positions (internal degrees of freedom)
        for each applied strain. Initial config `a` is not optimised, and should
        already have relaxed cell and atomic positions. Does not optimize atomic
        positions if set to None.
    verbose : bool
        If True, print additional infomation about the quality of fit
        and summarise results of C_ij and estimated errors. Default True.
    graphics : bool
        If True, use :mod:`matplotlib.pyplot` to plot the stress vs. strain
        curve for each C_ij component fitted. Default True. 
    logfile : bool
        Log file to write optimizer output to. Default None (i.e. suppress
        output).
    **kwargs : dict
        Additional arguments to pass to `optimizer.run()` method e.g. `fmax`.

    Returns
    -------
    C : array_like
        6x6 matrix of the elastic constants in Voigt notation.
    C_err : array_like
        If scipy.stats module is available then error estimates for each C_ij
        component are obtained from the accuracy of the linear regression.
        Otherwise an array of np.zeros((6,6)) is returned.
    
    Notes
    -----

    Code originally adapted from elastics.py script, available from
    http://github.com/djw/elastic-constants
    """

    if isinstance(a, Atoms):
        # we've been passed a single Atoms object: use it to generate
        # set of strained configurations according to symmetry
        strained_configs = generate_strained_configs(a, symmetry, N_steps, delta)
    else:
        # assume we've been passed a list of strained configs
        strained_configs = a

    if graphics:
        import matplotlib.pyplot as plt

    def do_fit(index1, index2, stress, strain, patt):
        if verbose:
            print('Fitting C_%d%d' % (index1+1, index2+1))
            print('Strain %r' % strain[:,index2])
            print('Stress %r GPa' % (stress[:,index1]/units.GPa))

        if scipy_stats is not None:
            cijFitted, intercept,r,tt,stderr = scipy_stats.linregress(strain[:,index2],stress[:,index1])
        else:
            cijFitted, intercept = np.polyfit(strain[:,index2],stress[:,index1], 1)
            r, tt, stderr = 0.0, None, 0.0

        if verbose:
            # print info about the fit
            print('Cij (gradient) / GPa    :    ',cijFitted/units.GPa)
            if scipy_stats is not None:
                print('Error in Cij / GPa      :    ', stderr/units.GPa)
                if abs(r) > 0.9:
                    print('Correlation coefficient :    ',r)
                else:
                    print('Correlation coefficient :    ',r, '     <----- WARNING')
            else:
                print('scipy.stats not available, no error estimation performed')

        if graphics:
            # position this plot in a 6x6 grid
            sp = plt.subplot(6,6,6*index1+(index2+1))
            sp.set_axis_on()

            # change the labels on the axes
            xlabels = sp.get_xticklabels()
            plt.setp(xlabels,'rotation',90,fontsize=7)
            ylabels = sp.get_yticklabels()
            plt.setp(ylabels,fontsize=7)

            # colour the plot depending on the strain pattern
            colourDict = {0: '#BAD0EF', 1:'#FFCECE', 2:'#BDF4CB', 3:'#EEF093',4:'#FFA4FF',5:'#75ECFD'}
            sp.set_axis_bgcolor(colourDict[patt])

            # plot the data
            plt.plot([strain[0,index2],strain[-1,index2]],
                     [(cijFitted*strain[0,index2]+intercept)/units.GPa,
                      (cijFitted*strain[-1,index2]+intercept)/units.GPa])
            plt.plot(strain[:,index2],stress[:,index1]/units.GPa,'ro')
            plt.xticks(strain[:,index2])

        return cijFitted, stderr

    if not symmetry in strain_patterns:
        raise ValueError('Unknown symmetry %s. Valid options are %s' % (symmetry, strain_patterns.keys()))

    # There are 21 independent elastic constants
    Cijs = {}
    Cij_err = {}

    # Construct mapping from (i,j) to index into Cijs in range 1..21
    # (upper triangle only to start with)
    Cij_map = {}
    Cij_map_sym = {}
    for i in range(6):
        for j in range(i,6):
            Cij_map[(i,j)] = Cij_symmetry[None][i,j]
            Cij_map_sym[(i,j)] = Cij_symmetry[symmetry][i,j]

    # Reverse mapping, index 1..21 -> tuple (i,j) with i, j in range 0..5
    Cij_rev_map = dict(zip(Cij_map.values(), Cij_map.keys()))

    # Add the lower triangle to Cij_map, e.g. C21 = C12
    for (i1,i2) in Cij_map.copy().keys():
        Cij_map[(i2,i1)] = Cij_map[(i1,i2)]
        Cij_map_sym[(i2,i1)] = Cij_map_sym[(i1,i2)]


    N_pattern = len(strain_patterns[symmetry])
    configs = iter(strained_configs)

    strain = np.zeros((N_pattern, N_steps, 6))
    stress = np.zeros((N_pattern, N_steps, 6))

    if graphics:
        fig = plt.figure(num=1, figsize=(9.5,8),facecolor='white')
        fig.clear()
        fig.subplots_adjust(left=0.07,right=0.97,top=0.97,bottom=0.07,wspace=0.5,hspace=0.5)

        for index1 in range(6):
            for index2 in range(6):
                # position this plot in a 6x6 grid
                sp = plt.subplot(6,6,6*(index1)+index2+1)
                sp.set_axis_off()
                plt.text(0.4,0.4, "n/a")

    # Fill in strain and stress arrays from config Atoms list
    for pattern_index, (pattern, fit_pairs) in enumerate(strain_patterns[symmetry]):
        for step in range(N_steps):
            at = next(configs)
            if optimizer is not None:
                optimizer(at, logfile=logfile).run(**kwargs)
            strain[pattern_index, step, :] = full_3x3_to_Voigt_6_strain(at.info['strain'])
            stress[pattern_index, step, :] = at.get_stress()

    # Do the linear regression
    for pattern_index, (pattern, fit_pairs) in enumerate(strain_patterns[symmetry]):
        for (index1, index2) in fit_pairs:
            fitted, err = do_fit(index1, index2,
                                 stress[pattern_index,:,:],
                                 strain[pattern_index,:,:],
                                 pattern_index)

            index = abs(Cij_map_sym[(index1, index2)])

            if not index in Cijs:
                if verbose:
                    print('Setting C%d%d (%d) to %f +/- %f' %
                          (index1+1, index2+1, index, fitted, err))
                Cijs[index] = [fitted]
                Cij_err[index] = [err]
            else:
                if verbose:
                    print('Updating C%d%d (%d) with value %f +/- %f' %
                          (index1+1, index2+1, index, fitted, err))
                Cijs[index].append(fitted)
                Cij_err[index].append(err)
            if verbose:
                print('\n')


    C = np.zeros((6,6))
    C_err = np.zeros((6,6))
    C_labels = np.zeros((6,6),dtype='S4')
    C_labels[:] = '    '

    # Convert lists to mean
    for k in Cijs:
        Cijs[k] = np.mean(Cijs[k])

    # Combine statistical errors
    for k, v in Cij_err.items():
        Cij_err[k] = np.sqrt(np.sum(np.array(v)**2))/np.sqrt(len(v))

    if symmetry.startswith('trigonal'):
        # Special case for trigonal lattice: C66 = (C11 - C12)/2
        Cijs[Cij_map[(5,5)]] = 0.5*(Cijs[Cij_map[(0,0)]]-Cijs[Cij_map[(0,1)]])
        Cij_err[Cij_map[(5,5)]] = np.sqrt(Cij_err[Cij_map[(0,0)]]**2 + Cij_err[Cij_map[(0,1)]]**2)

    # Generate the 6x6 matrix of elastic constants
    # - negative values signify a symmetry relation
    for i in range(6):
        for j in range(6):
            index = Cij_symmetry[symmetry][i,j]
            if index > 0:
                C[i,j] = Cijs[index]
                C_err[i,j] = Cij_err[index]
                ii, jj = Cij_rev_map[index]
                C_labels[i,j] = ' C%d%d' % (ii+1,jj+1)
                C_err[i,j] = Cij_err[index]
            elif index < 0:
                C[i,j] = -Cijs[-index]
                C_err[i,j] = Cij_err[-index]
                ii, jj = Cij_rev_map[-index]
                C_labels[i,j] = '-C%d%d' % (ii+1, jj+1)

    if verbose:
        print(np.array2string(C_labels).replace("'",""))
        print('\n = \n')
        print(np.array2string(C/units.GPa,
                              suppress_small=True,
                              precision=2))
        print

        # Summarise the independent components of C_ij matrix
        printed = {}
        for i in range(6):
            for j in range(6):
                index = Cij_symmetry[symmetry][i,j]
                if index <= 0 or index in printed: continue
                print('C_%d%d = %-4.2f +/- %-4.2f GPa' %
                      (i+1, j+1, C[i,j]/units.GPa, C_err[i,j]/units.GPa))
                printed[index] = 1

    return C, C_err


def youngs_modulus(C, l):
    """
    Calculate approximate Youngs modulus E_l from 6x6 elastic constants matrix C_ij

    This is the modulus for loading in the l direction. For the exact answer, taking
    into account elastic anisotropuy, rotate the C_ij matrix to the correct frame,
    compute the compliance matrix::

       C = ...  # 6x6 C_ij matrix in crystal frame
       A = ...  # rotation matrix
       Cr = rotate_elastic_constants(C, A)
       S = np.inv(Cr)
       E_x = 1/S[0, 0]  # Young's modulus for a pull in x direction
       E_y = 1/S[1, 1]  # Young's modulus for a pull in y direction
       E_z = 1/S[0, 0]  # Young's modulus for a pull in z direction

    Notes
    -----

    Formula is from W. Brantley, Calculated elastic constants for stress problems associated 
    with semiconductor devices. J. Appl. Phys., 44, 534 (1973).
    """  

    S = inv(C)        # Compliance matrix
    lhat = l/norm(l)  # Normalise directions

    # Youngs modulus in direction l, ratio of stress sigma_l 
    # to strain response epsilon_l
    E = 1.0/(S[0,0] - 2.0*(S[0,0]-S[0,1]-0.5*S[3,3])*(lhat[0]*lhat[0]*lhat[1]*lhat[1] +
         lhat[1]*lhat[1]*lhat[2]*lhat[2] +
         lhat[0]*lhat[0]*lhat[2]*lhat[2]))
    return E


def poisson_ratio(C, l, m):
    """
    Calculate approximate Poisson ratio \nu_{lm} from 6x6 elastic constant matrix C_{ij}

    This is the response in `m` direction to pulling in `l` direction. Result is dimensionless.

    Notes
    -----

    Formula is from W. Brantley, Calculated elastic constants for stress problems associated 
    with semiconductor devices. J. Appl. Phys., 44, 534 (1973).
    """
    
    S = inv(C)        # Compliance matrix
    lhat = l/norm(l)  # Normalise directions
    mhat = m/norm(m)

    # Poisson ratio v_lm: response in m direction to strain in 
    # l direction, v_lm = - epsilon_m/epsilon_l
    v = -((S[0,1] + (S[0,0]-S[0,1]-0.5*S[3,3])*(lhat[0]*lhat[0]*mhat[0]*mhat[0] +
         lhat[1]*lhat[1]*mhat[1]*mhat[1] +
         lhat[2]*lhat[2]*mhat[2]*mhat[2])) / 
         (S[0,0] - 2.0*(S[0,0]-S[0,1]-0.5*S[3,3])*(lhat[0]*lhat[0]*lhat[1]*lhat[1] + 
         lhat[1]*lhat[1]*lhat[2]*lhat[2] + 
         lhat[0]*lhat[0]*lhat[2]*lhat[2])))
    return v


def elastic_moduli(C, l=np.array([1, 0, 0]), R=None, tol=1e-6):
    """
    Calculate elastic moduli from 6x6 elastic constant matrix C_{ij}.
    
    The elastic moduli calculated are: Young's muduli, Poisson's ratios,
    shear moduli, bulk mudulus and bulk mudulus tensor.
    
    If a direction l is specified, the system is rotated to have it as its
    x direction (see Notes for details). If R is specified the system is
    rotated according to it.
    
    Parameters
    ----------
    C : array_like
        6x6 stiffness matrix (Voigt notation).
    l : array_like, optional
        3D direction vector for pull (the default is the x direction
        of the original system)
    R : array_like, optional
        3x3 rotation matrix.
    
    Returns
    -------
    E : array_like
        Young's modulus for a stress in each of the three directions
        of the rotated system.
    nu : array_like
         3x3 matrix with Poisson's ratios.
    Gm : array_like
         3x3 matrix with shear moduli.
    B : float
        Bulk modulus.
    K : array_like
        3x3 matrix with bulk modulus tensor.
    
    Other Parameters
    ----------------
    tol : float, optional
        tolerance for checking validity of rotation and comparison
        of vectors.
    
    Notes
    ---
    It works by rotating the elastic constant tensor to the desired
    direction, so it should be valid for arbitrary crystal structures.
    If only l is specified there is an infinite number of possible
    rotations. The chosen one is a rotation along the axis orthogonal
    to the plane defined by the vectors (1, 0, 0) and l.
    
    Bulk modulus tensor as defined in
    O. Rand and V. Rovenski, "Analytical Methods in Anisotropic
    Elasticity", Birkh\"auser (2005), pp. 71.
    
    """
    
    if R is not None:
        R = np.asarray(R)

        # Is this a rotation matrix?
        if np.sometrue(np.abs(np.dot(np.array(R), np.transpose(np.array(R))) - 
                              np.eye(3, dtype=float)) > tol):
            raise RuntimeError('Matrix *R* does not describe a rotation.')
    else:
        u_a = np.array([1, 0, 0])
        u_b = l/norm(l)  # Normalise directions
        R = np.eye(3)

        if not np.allclose(l, u_a, rtol=tol, atol=tol):
            u_v = np.cross(u_a, u_b)
            u_v_mat = np.array([[ 0,      -u_v[2], u_v[1]],
                                [ u_v[2], 0,      -u_v[0]],
                                [-u_v[1], u_v[0], 0]])

            R = R + u_v_mat + \
                np.dot(u_v_mat, u_v_mat) * (1 - np.dot(u_a, u_b)) / \
                np.linalg.norm(u_v)**2

    Cr = rotate_elastic_constants(C, R)
    S = np.linalg.inv(Cr)

    # Young's modulus for a stress in \alpha direction; \alpha = x, y, z
    E = np.zeros(3)
    E[0] = 1/S[0, 0]
    E[1] = 1/S[1, 1]
    E[2] = 1/S[2, 2]

    # Poisson's ratio ($\alpha$, $\beta$); $\alpha$, $\beta$ = x, y, z
    nu = np.array([
            [-1,               -S[1, 0]/S[0, 0], -S[2, 0]/S[0, 0]],
            [-S[0, 1]/S[1, 1], -1,               -S[2, 1]/S[1, 1]],
            [-S[0, 2]/S[2, 2], -S[1, 2]/S[2, 2], -1]
            ])

    # Shear modulus ($\alpha$, $\beta$); $\alpha$, $\beta$ = x, y, z
    G = np.zeros(3)
    G[0] = 1/S[3, 3]  # Shear modulus yz
    G[1] = 1/S[4, 4]  # Shear modulus zx
    G[2] = 1/S[5, 5]  # Shear modulus xy
    Gm = np.array([
            [E[0]/4,  G[2],    G[1]],
            [G[2],    E[1]/4,  G[0]],
            [G[1],    G[0],    E[2]/4]
            ])

    # Bulk modulus
    B = 1/np.sum(S[0:3, 0:3])
    
    # Bulk modulus tensor
    Crt = Voigt_6x6_to_full_3x3x3x3(Cr)
    K = np.einsum('ijkk', Crt)

    return E, nu, Gm, B, K



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


class BaseQMClusterTool(object):
    """Class that hold all the functions common to
    qm clustering objects"""

    def __init__(self, mediator):
        """Class that hold all the functions common to
        qm clustering objects

        Parameters
        ----------
        mediator : matscipy.calculators.mcfm.QMCluster
            class responsible for managing the QM clusters in the simulation
        """
        self.mediator = mediator
        self.verbose = mediator.verbose

    def find_neighbours(self, atoms, index):
        """Find the neighbours of atom i using self.neighbour_list
        returns a list of [heavy_neighbours, hydrogen_neighbours]

        Parameters
        ----------
        atoms : ase.Atoms object
            structure in which it is necessary to find the neighbours
        index : int
            atomic index

        Returns
        -------
        list
            non-hydrogen neighbours
        list
            hydrogen neighbours
        """

        neighbours = self.mediator.neighbour_list.get_neighbours(index)

        heavy_n = []
        hydro_n = []
        for arr_i, atom_i in enumerate(neighbours):
            if (atoms.numbers[atom_i] == 1):
                hydro_n.append(atom_i)
            else:
                heavy_n.append(atom_i)

        return [heavy_n, hydro_n]

    def hydrogenate_cluster(self, atoms, cluster):
        """Add neigoburing hydrogens to a cluster composed of heavy ions
        The input should be a set representing heavy ions in a cluster
        This functions operates on sets

        Parameters
        ----------
        atoms : ase.Atoms object
            structure in which it is necessary to find the neighbours
        cluster :ase.Atoms object
            sub-structure of the larger struct that needs its dangling
            bonds hydrogenated

        Returns
        -------
        ase.Atoms
            The original cluster but now hydrogenated
        """

        for atom_id in cluster.copy():
                # find_neighbours returns a list where
                # [0] - heavy neighbours
                # [1] - hydrogen neighbours
            cluster |= set(self.find_neighbours(atoms, atom_id)[1])

        return cluster
